from datetime import datetime, timedelta
from office365.runtime.auth.client_credential import ClientCredential
from office365.sharepoint.client_context import ClientContext
from django.conf import settings


class SharePointContext:
    _instance = None
    _client_credentials = None
    _ctx = None
    _last_created = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(SharePointContext, cls).__new__(cls)
            cls._initialize(cls._instance)
        elif datetime.now() - cls._last_created > timedelta(hours=23):
            cls._initialize(cls._instance)
        return cls._instance

    @classmethod
    def _initialize(cls, instance):
        client_id = getattr(settings, 'SHAREPOINT_APP_CLIENT_ID', 'client_id')
        sharepoint_url = getattr(settings, 'SHAREPOINT_URL', 'sharepoint_url')
        sharepoint_api_certificate_path = getattr(settings, 'SHAREPOINT_API_CERTIFICATE_PATH', 'sharepoint_api_certificate_path')
        sharepoint_api_certificate_thumbprint = getattr(settings, 'SHAREPOINT_API_CERTIFICATE_THUMBPRINT', 'sharepoint_api_certificate_thumbprint')
        sharepoint_api_tenant_name = getattr(settings, 'SHAREPOINT_API_TENANT_NAME', 'sharepoint_api_tenant_name')

        cert_settings = {
            'client_id': client_id,
            'thumbprint': sharepoint_api_certificate_thumbprint,
            'cert_path': sharepoint_api_certificate_path
        }
        instance._ctx = ClientContext(sharepoint_url).with_client_certificate(sharepoint_api_tenant_name, **cert_settings)
        cls._last_created = datetime.now()

    @property
    def client_credentials(self):
        return self._client_credentials

    @property
    def ctx(self):
        return self._ctx


from pathlib import Path
from urllib.parse import urlparse
from django.conf import settings
from django_sharepoint_storage.SharePointStorage import SharePointStorage

Static = lambda: SharePointStorage(location=getattr(settings, 'SHAREPOINT_STATIC_DIR', 'sharepoint_static_dir'))
Media = lambda: SharePointStorage(location=getattr(settings, 'SHAREPOINT_MEDIA_DIR', 'sharepoint_media_dir'))

def get_server_relative_path(url):
    parse_result = urlparse(url)
    path_parts = Path(parse_result.path).parts
    if path_parts:
        return '/'.join(path_parts)[1:]
    else:
        return ''


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    name="DjangoSharepointStorage",
    version="1.1.2",
    author="Melih Sünbül",
    author_email="m.sunbul@lund-it.com",
    description="A Python library to use SharePoint as storage backend for your Django application",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LundIT/DjangoSharepointStorage",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "django",
        "Office365-REST-Python-Client",
    ],
    python_requires='>=3.6',
)



import platform
from django.core.files.storage import Storage
import os
import datetime
import time
from io import BytesIO
from django.conf import settings
from django.db import connection

from django_sharepoint_storage.SharePointContext import SharePointContext
from django_sharepoint_storage.SharePointFile import SharePointFile

DB_NAME = connection.settings_dict['NAME']


class SharePointStorage(Storage):
    sharepoint_url = getattr(settings, 'SHAREPOINT_URL', 'sharepoint_url')

    def __init__(self, location='uploads', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.location = location

    @staticmethod
    def print_failure(retry_number, ex):
        shrp_ctx = SharePointContext()
        print(f"{retry_number}: {ex}")
        if retry_number == 5:
            shrp_ctx.ctx._queries.pop()
            raise ex

    def _open(self, name, mode='rb', retries=5):
        from django_sharepoint_storage.SharePointCloudStorageUtils import get_server_relative_path

        if mode in ['r', 'rb']:
            file_url = self.url(name)
            shrp_ctx = SharePointContext()
            file = shrp_ctx.ctx.web.get_file_by_server_relative_path(
                get_server_relative_path(file_url)).execute_query_retry(max_retry=5, timeout_secs=5,
                                                                        failure_callback=SharePointStorage.print_failure)
            if retries <= 0:
                raise Exception("SharePoint Server cannot handle requests at the moment.")
            try:
                binary_file = file.open_binary(shrp_ctx.ctx, get_server_relative_path(file_url))
                bytesio_object = BytesIO(binary_file.content)
                return bytesio_object
            except Exception as ex:
                if ex.response.status_code == 404:
                    raise ex
                time.sleep(5)
                return self._open(name, mode, retries - 1)

        elif mode in ['w', 'wb', "w+", "wb+"]:
            return SharePointFile(name, mode, self)
        else:
            raise ValueError(f"Unsupported file mode: {mode}")

    def _save(self, name, content):
        shrp_ctx = SharePointContext()
        folder_path = f"Shared Documents/{os.getenv('DEPLOYMENT_ENVIRONMENT', 'LOCAL')}-{os.getenv('K8S_NAMESPACE', 'ENV')}/{os.getenv('KEYCLOAK_INTERNAL_CLIENT_ID', 'Local')}/{os.getenv('INSTANCE_RESOURCE_IDENTIFIER', f'{platform.node()}/{DB_NAME}')}/{self.location}/{os.path.dirname(name)}"
        target_folder = shrp_ctx.ctx.web.ensure_folder_path(folder_path).get().select(
            ["ServerRelativePath"]).execute_query_retry(max_retry=5, timeout_secs=5,
                                                        failure_callback=SharePointStorage.print_failure)
        content.seek(0)
        file_content = content.read()

        target_folder.upload_file(os.path.basename(name), file_content).execute_query_retry(max_retry=5, timeout_secs=5,
                                                                                            failure_callback=SharePointStorage.print_failure)

        return name

    def delete(self, name):
        from django_sharepoint_storage.SharePointCloudStorageUtils import get_server_relative_path

        shrp_ctx = SharePointContext()

        file_path = get_server_relative_path(self.url(name))
        file = shrp_ctx.ctx.web.get_file_by_server_relative_path(file_path).get().execute_query_retry(max_retry=5,
                                                                                                      timeout_secs=5,
                                                                                                      failure_callback=SharePointStorage.print_failure)
        file.delete_object().execute_query_retry(max_retry=5, timeout_secs=5,
                                                 failure_callback=SharePointStorage.print_failure)

    def exists(self, name, retries=5):
        from django_sharepoint_storage.SharePointCloudStorageUtils import get_server_relative_path

        file_path = get_server_relative_path(self.url(name))

        shrp_ctx = SharePointContext()

        if retries <= 0:
            raise Exception("SharePoint Server cannot handle requests at the moment.")
        try:
            if name.endswith('/'):
                shrp_ctx.ctx.web.get_folder_by_server_relative_path(file_path).get().execute_query()
            else:
                shrp_ctx.ctx.web.get_file_by_server_relative_path(file_path).get().execute_query()
        except Exception as ex:
            if ex.response.status_code == 404:
                return False
            time.sleep(5)
            return self.exists(name, retries - 1)

        return True

    def get_modified_time(self, name):
        from django_sharepoint_storage.SharePointCloudStorageUtils import get_server_relative_path

        file_path = get_server_relative_path(self.url(name))

        shrp_ctx = SharePointContext()

        if name.endswith('/'):
            file = shrp_ctx.ctx.web.get_folder_by_server_relative_path(file_path).get().expand(
                ["TimeLastModified"]).get().execute_query_retry(max_retry=5, timeout_secs=5,
                                                                failure_callback=SharePointStorage.print_failure)
        else:
            file = shrp_ctx.ctx.web.get_file_by_server_relative_path(file_path).get().expand(
                ["TimeLastModified"]).get().execute_query_retry(max_retry=5, timeout_secs=5,
                                                                failure_callback=SharePointStorage.print_failure)

        return datetime.datetime.timestamp(file.time_last_modified)

    def listdir(self, name):
        from django_sharepoint_storage.SharePointCloudStorageUtils import get_server_relative_path

        shrp_ctx = SharePointContext()

        folder_path = get_server_relative_path(self.url(name))

        root_folder = shrp_ctx.ctx.web.get_folder_by_server_relative_path(folder_path).get().execute_query_retry(
            max_retry=5, timeout_secs=5, failure_callback=SharePointStorage.print_failure)
        files = root_folder.get_files(recursive=False).execute_query_retry(max_retry=5, timeout_secs=5,
                                                                           failure_callback=SharePointStorage.print_failure)
        return [f.name for f in root_folder.folders], [f.name for f in files]

    def url(self, name):
        # Use the dirname of name as your upload_to equivalent
        return f"{self.sharepoint_url}/Shared Documents/{os.getenv('DEPLOYMENT_ENVIRONMENT', 'LOCAL')}-{os.getenv('K8S_NAMESPACE', 'ENV')}/{os.getenv('KEYCLOAK_INTERNAL_CLIENT_ID', 'Local')}/{os.getenv('INSTANCE_RESOURCE_IDENTIFIER', f'{platform.node()}/{DB_NAME}')}/{self.location}/{name}"
